<!DOCTYPE html>
<html lang="en">
<head>
    <meta charset="UTF-8">
    <title>Training ML model - CV details</title>
    <style>
        {{{css_style}}}
    </style>
    <meta name="viewport" content="width=device-width, initial-scale=1">
</head>
<body>
    <div class="container padded-bottom">
        <h3>Assessment split {{split_index}} details: label {{label}}, hyperparameter setting {{hp_setting}}</h3>
        {{#has_ml_reports}}
        <h3>ML reports</h3>
        {{#ml_reports}}
        <h4>Report {{name}}</h4>
        <p>{{info}}</p>
        <div class="grid-container">
            {{#output_figures}}
            <div class="">
                {{#is_embed}}
                <embed src="{{path}}">
                {{/is_embed}}
                {{^is_embed}}
                <img src="{{path}}">
                {{/is_embed}}
                <p>{{name}}</p>
            </div>
            {{/output_figures}}
        </div>
        <ul>
        {{#output_tables}}
        <li><a href="{{path}}">Download {{#name}}{{name}} report{{/name}}</a></li>
        {{/output_tables}}
        </ul>

        {{/ml_reports}}
        {{/has_ml_reports}}

        {{#has_encoding_reports}}
        <h3>Reports on encoded datasets</h3>
        <div class="grid-container padded-top padded-bottom">
            <div class="">
                {{#encoding_train_reports}}
                <h4>Report {{name}} (training)</h4>
                <p>{{info}}</p>
                {{#output_figures}}
                {{#is_embed}}
                <embed src="{{path}}">
                {{/is_embed}}
                {{^is_embed}}
                <img src="{{path}}">
                {{/is_embed}}
                {{#name}}<p>{{name}} (training data)</p>{{/name}}
                {{/output_figures}}

                <ul>
                {{#output_tables}}
                <li><a href="{{path}}">Download {{#name}}{{name}} (training data){{/name}}</a></li>
                {{/output_tables}}
                </ul>
                {{/encoding_train_reports}}
            </div>
            <div class="">
                {{#encoding_test_reports}}
                <h4>Report {{name}} (test)</h4>
                <p>{{info}}</p>
                {{#output_figures}}
                {{#is_embed}}
                <embed src="{{path}}">
                {{/is_embed}}
                {{^is_embed}}
                <img src="{{path}}">
                {{/is_embed}}
                {{#name}}<p>{{name}} (test data)</p>{{/name}}
                {{/output_figures}}

                <ul>
                {{#output_tables}}
                <li><a href="{{path}}">Download {{#name}}{{name}} (test data){{/name}}</a></li>
                {{/output_tables}}
                </ul>
                {{/encoding_test_reports}}
            </div>
        </div>
        {{/has_encoding_reports}}
    </div>
</body>
</html>