<!DOCTYPE html>
<html lang="en">
<head>
    <meta charset="UTF-8">
    <title>immuneML: Result Overview</title>
    <link rel="stylesheet" type="text/css" href="{{{css_path}}}">
    <meta name="viewport" content="width=device-width, initial-scale=1">

</head>
<body>
    <div class="container">
        <h3>Results of instructions</h3>
        <div>Full specification is available <a href="{{{full_specs}}}">here</a>.</div>
        <div class="padded-bottom padded-top">immuneML version: {{immuneML_version}}</div>
        {{#instructions}}
            <h4>{{instruction_class}}: {{instruction_name}}</h4>
            <a href="{{path}}">Report link</a>
        {{/instructions}}
    </div>
</body>
</html>